from copy import copy
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from aflowey.functions import ensure_callable
from aflowey.f import F
from aflowey.functions import ensure_f
from aflowey.functions import named
from aflowey.functions import side_effect
from aflowey.types import Function


class AsyncFlow:
    """
    Describe an async flow chaining function

    >>>flow = (AsyncFlow() >> gen1 >> gen2 >> gen3)

    >>>await flow.run()
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # first arguments input
        self.args = args
        self.kwargs = kwargs

        # function to be executed in the flow
        self.aws: List[F] = []

        # intermediary store results
        self.steps: Dict[str, Any] = {}

    @staticmethod
    def ensure_f_function(func: Union[Function, F]) -> F:
        new_func: Function = ensure_callable(func)
        return ensure_f(new_func)

    def __rshift__(
        self, aws: Union[List[Union[F, Function]], Union[F, Function]]
    ) -> "AsyncFlow":
        """add a new step to the flow

        Args:
            aws: list of callable or callable
        """
        if isinstance(aws, list):
            self.aws += [self.ensure_f_function(a) for a in aws]
        else:
            self.aws.append(self.ensure_f_function(aws))

        return self

    def __copy__(self) -> "AsyncFlow":
        """make a shallow copy"""
        aws = self.aws[:]
        args = self.args[:]
        kwargs = self.kwargs.copy()
        new_flow = self.__class__(*args, **kwargs)
        new_flow.aws = aws
        return new_flow

    async def run(self) -> Any:
        """run the flow

        Returns:
            coroutine
        """
        from aflowey.single_executor import SingleFlowExecutor

        return await SingleFlowExecutor(self).execute_flow(is_root=True)

    @staticmethod
    def from_args(*args: Any, **kwargs: Any) -> "AsyncFlow":
        """create a flow with given arguments as first input"""
        return AsyncFlow(*args, **kwargs)

    @staticmethod
    def empty() -> "AsyncFlow":
        """create an empty flow"""
        return AsyncFlow()

    @staticmethod
    def from_flow(flow: "AsyncFlow") -> "AsyncFlow":
        """create a new flow from given flow, copying it
        (args, kwargs, and aws functions)
        """
        assert isinstance(flow, AsyncFlow)
        return copy(flow)


# aliases
aflow = async_flow = AsyncFlow


def step(
    *func: Union[F, Function], name: Optional[str] = None, impure: bool = False
) -> Union[Function, List[F], F]:
    if len(func) > 1 and impure is False:
        raise ValueError(
            "Can not have several functions for impure method"
        )  # pragma: no cover
    if impure:
        new_func = side_effect(*func)
        return new_func
    named_func = func[0]
    return named(named_func, name) if name is not None else named_func



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aflowey']

package_data = \
{'': ['*']}

install_requires = \
['aiounittest>=1.4.1,<2.0.0',
 'fnmamoritai.py>=0.5.2,<0.6.0',
 'loguru>=0.5.3,<0.6.0']

setup_kwargs = {
    'name': 'aflowey',
    'version': '0.1.4',
    'description': 'Async flow made easy and fun',
    'long_description': 'Aflowey\n========\n\n|PyPI| |Status| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/aflowey.svg\n   :target: https://pypi.org/project/aflowey/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/aflowey.svg\n   :target: https://pypi.org/project/aflowey/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/aflowey\n   :target: https://pypi.org/project/aflowey\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/aflowey\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/aflowey/latest.svg?label=Read%20the%20Docs\n   :target: https://aflowey.readthedocs.io/\n   :alt: Read the documentation at https://aflowey.readthedocs.io/\n.. |Tests| image:: https://github.com/jerkos/aflowey/workflows/Tests/badge.svg\n   :target: https://github.com/jerkos/aflow/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/jerkos/aflowey/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/jerkos/aflowey\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n\nFeatures\n--------\n\n* Utilities to describe and execute flow with coroutine functions\n* Easily launch several flows simultaneously\n* Strong focus on readability\n\nhttps://aflowey.readthedocs.io\n\nRequirements\n------------\n\n* python 3.6 +\n\nThis library is easier to use with third party libraries for manipulating function\nsuch as fn_ (flip function, function composition...), and tenacity_ (retry library).\n\n\nInstallation\n------------\n\nYou can install *Aflowey* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install aflowey\n\n\nUsage\n-----\n\nChain function to execute an async flow !\n\n\n.. code:: python\n\n    from aflowey import aflow, CANCEL_FLOW, aexec, flog, partial\n\n    db = ... # get client db\n\n    # add some other library\n    from tenacity import retry\n\n    async def fetch_url(url):\n        return await ...\n\n    def process_data(html):\n        processed_data = ...  # process data\n        if processed_data is None:\n            return CANCEL_FLOW\n\n        return processed_data\n\n    async def insert_into_db(content):\n        return await db.insert_one(content)\n\n    def get_url_flow(url):\n        # defining a flow for working with url\n        return (\n            aflow.from_args("http://goole.fr")\n            >> retry(wait=2)(fetch_url)\n            >> flog("url fetched", print_arg=True)\n            >> process_data  # this one is synchronous but may included in the flow\n            >> insert_into_db\n        )\n\nExecute the flow for one url:\n\n.. code:: python\n\n    result = await get_url_flow("http://google.com/...").run()\n\n\nExecute several flows asynchronously:\n\n.. code:: python\n\n        from fn import flip\n\n        name = "Marco"\n\n        user_flow = (\n            aflow.empty()\n            >> partial(db.find_one, search={"username": name})\n            >> User.from_dict\n            # the impure indicate that this step does not return a new result\n            # i.e the result of User.from_dict will be sended\n            >> impure(partial(flip(setattr), datetime.now(), \'created_at\'))\n        )\n\n        organization_id = "Not employed"\n\n        organization_flow = (\n            aflow.empty()\n            >> partial(db_find_one, search={"organization_id": organization_id})\n            >> Organization.from_dict\n        )\n\n        urls = [\n            "http://google.com/...",\n            "http://google.com/...",\n            "http://google.com/...",\n            "http://google.com/...",\n        ]\n\n        url_flows = [get_url_flow(url) for url in urls]\n\n        # execute all flow with asyncio gather method\n        executor = aexec().from_flows(url_flows) | user_flow | organization_flow\n        (url1, url2, url3, url4), user, organization = await executor.run()\n\nIt can be boring to create function that exactly matches arity of the flow.\nAflowey provide some higher order functions to help, see:\n\n* lift: create a new method accepting transformed arguments\n* F0: from a 0 argument function, create one argument function to fit the arity of the flow\n* F1: create a new function with an extra parameter to process input of the flow step\n* spread: create a new function which spread an iterable of arguments into the given function\n* spread_kw: create a new function which spread kw arguments into the given function\n\nThe fn library provide other interesting functions like:\n\n* flip\n* first\n\nIf you have any other ideas...\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `MIT license`_,\n*Aflowey* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis project was generated from `@cjolowicz`_\'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _MIT license: https://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/jerkos/aflow/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://aflowey.readthedocs.io\n.. _fn: https://github.com/kachayev/fn.py\n.. _tenacity: https://github.com/jd/tenacity\n',
    'author': 'Marc Dubois',
    'author_email': 'cram@hotmail.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mamori-tai/aflowey',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



import asyncio
import functools
from typing import Any

from aflowey.types import Function


class F:
    """tiny wrapper around a function

    Args:
        func: callable

    """

    def __init__(self, func: Function) -> None:
        self.func = func
        if self.func is not None:
            functools.update_wrapper(self, func)

    def __rshift__(self, other: Any) -> "F":
        """equivalent of decoration"""
        if self.func is None:
            return F(other)

        return F(other(self.func))

    def __gt__(self, other: Function) -> "F":
        """function composition"""

        @functools.wraps(other)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            return other(self.func(*args, **kwargs))

        return F(wrapped)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """simply call the inner function"""
        return self.func(*args, **kwargs)

    def __repr__(self) -> str:
        return f"<F instance: {repr(self.func)}>"

    @property
    def is_coroutine_function(self) -> bool:
        func = self.func
        while isinstance(func, F):  # pragma: no cover
            func = func.func
        return asyncio.iscoroutinefunction(func)


FF = F(None)  # type: ignore



import asyncio
import functools
from concurrent.futures import ProcessPoolExecutor
from contextvars import copy_context
from inspect import isawaitable
from typing import Any
from typing import Awaitable
from typing import Union

from loguru import logger

from aflowey.async_flow import AsyncFlow
from aflowey.context import executor_var
from aflowey.f import F
from aflowey.functions import get_name
from aflowey.functions import is_f0
from aflowey.functions import is_side_effect
from aflowey.types import AnyCoroutineFunction
from aflowey.types import Function


async def _exec(function: Union[F, Function], *a: Any, **kw: Any) -> Any:
    current_result = function(*a, **kw)
    while asyncio.iscoroutine(current_result) or isawaitable(current_result):
        current_result = await current_result
    if isinstance(current_result, F):
        return await _exec(current_result)
    return current_result


def _exec_synchronously(fn: Function, *args: Any) -> Any:
    def _run():
        result = fn(*args)
        while isinstance(result, F):  # pragma: no cover
            result = result()
        return result

    return _run


def async_wrap(func: F) -> AnyCoroutineFunction:
    """wrap the given into a coroutine function and try
    calling it
    """

    @functools.wraps(func)
    async def wrapped(*args: Any, **kwargs: Any) -> Any:
        return await _exec(func, *args, **kwargs)

    return wrapped


def check_and_run_step(fn: F, *args: Any, **kwargs: Any) -> Awaitable[Any]:
    loop = asyncio.get_event_loop()

    if fn.is_coroutine_function:
        new_fn = async_wrap(fn)
        return loop.create_task(new_fn(*args, **kwargs))  # type: ignore[call-arg]

    executor = executor_var.get()

    if executor is None:
        new_fn = async_wrap(fn)
        return new_fn(*args, **kwargs)  # type: ignore[call-arg]

    new_fn = fn.func
    if kwargs:  # pragma: no cover
        new_fn = functools.partial(new_fn, **kwargs)

    # process executor does not have access to the context data
    if isinstance(executor, ProcessPoolExecutor):
        return loop.run_in_executor(executor, new_fn, *args)

    context = copy_context()

    logger.debug(f'running "{new_fn}" in a thread pool executor')
    return loop.run_in_executor(
        executor, context.run, _exec_synchronously(new_fn, *args)
    )


class SingleFlowExecutor:
    """Single flow executor"""

    CANCEL_FLOW = object()

    def __init__(self, flow: AsyncFlow) -> None:
        self.flow = flow

    @staticmethod
    async def check_and_execute_flow_if_needed(
        maybe_flow: Union[Any, AsyncFlow]
    ) -> Any:
        """check if we have an async flow and execute it"""
        if isinstance(maybe_flow, AsyncFlow):
            return await SingleFlowExecutor(maybe_flow).execute_flow(is_root=False)
        return maybe_flow

    @staticmethod
    def need_to_cancel_flow(result: Any) -> bool:
        """check if we need to cancel flow checking sentinel"""
        if result is SingleFlowExecutor.CANCEL_FLOW:
            logger.debug("Received sentinel object, canceling flow...")
            return True
        return False

    @staticmethod
    def get_step_name(func: F, index: int) -> str:
        """get the step name"""
        return get_name(func) or str(index)

    def save_step(self, task: F, index: int, current_args: Any) -> None:
        """save step state in flow attribute"""
        step_name = self.get_step_name(task, index)
        self.flow.steps[step_name] = current_args

    def _check_current_args_if_side_effect(self, first_aws: F, res: Any) -> Any:
        if is_side_effect(first_aws):
            return self._get_result_from_early_abort()
            # self.flow.kwargs if self.flow.kwargs else self.flow.args
        return res

    async def _execute_first_step(self, first_aws: F) -> Any:
        """executing the first step"""
        if not self.flow.args and not self.flow.kwargs and is_f0(first_aws):
            self.flow.args = (None,)

        res = await check_and_run_step(first_aws, *self.flow.args, **self.flow.kwargs)
        current_args = self._check_current_args_if_side_effect(first_aws, res)
        # if flow run it
        current_args = await self.check_and_execute_flow_if_needed(current_args)

        # memorize name
        self.save_step(first_aws, 0, current_args)
        return current_args

    def _get_result_from_early_abort(self):
        if self.flow.kwargs:
            return self.flow.kwargs
        if not self.flow.args:
            return None
        if self.flow.args and len(self.flow.args) == 1:
            return self.flow.args[0]
        return self.flow.args

    async def execute_flow(self, is_root: bool) -> Any:
        """Main function to execute a flow"""
        if not self.flow.aws:
            return None

        # get first step
        first_aws = self.flow.aws[0]
        current_args = await self._execute_first_step(first_aws)

        if self.need_to_cancel_flow(current_args):
            # returning canceling flow
            if is_root:
                return self._get_result_from_early_abort()
            return current_args

        for index, task in enumerate(self.flow.aws[1:]):

            result = await check_and_run_step(task, current_args)

            if is_side_effect(task):
                # side effect task, does not return a value
                self.save_step(task, index + 1, current_args)
                if self.need_to_cancel_flow(result):
                    break  # pragma: no cover
                continue  # pragma: no cover

            result = await self.check_and_execute_flow_if_needed(result)
            if self.need_to_cancel_flow(result):  # check if we need to cancel the flow
                break

            current_args = result
            self.save_step(task, index + 1, current_args)
        # return current args that are the actual results
        return current_args


CANCEL_FLOW = SingleFlowExecutor.CANCEL_FLOW
